/* -*- Mode: C; indent-tabs-mode: f; c-basic-offset: 4; tab-width: 4 -*- */

/* nemo-global-preferences.c - Nemo specific preference keys and
                                   functions.

   Copyright (C) 1999, 2000, 2001 Eazel, Inc.

   The Gnome Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   The Gnome Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with the Gnome Library; see the file COPYING.LIB.  If not,
   write to the Free Software Foundation, Inc., 51 Franklin Street - Suite 500,
   Boston, MA 02110-1335, USA.

   Authors: Ramiro Estrugo <ramiro@eazel.com>
*/

#include <config.h>
#include "nemo-global-preferences.h"

#include "nemo-file-utilities.h"
#include "nemo-file.h"
#include <eel/eel-glib-extensions.h>
#include <eel/eel-gtk-extensions.h>
#include <eel/eel-stock-dialogs.h>
#include <eel/eel-string.h>
#include <eel/eel-debug.h>
#include <glib/gi18n.h>
#include <gio/gdesktopappinfo.h>

GSettings *nemo_preferences;
GSettings *nemo_icon_view_preferences;
GSettings *nemo_list_view_preferences;
GSettings *nemo_compact_view_preferences;
GSettings *nemo_desktop_preferences;
GSettings *nemo_tree_sidebar_preferences;
GSettings *nemo_window_state;
GSettings *nemo_plugin_preferences;
GSettings *nemo_menu_config_preferences;
GSettings *gnome_lockdown_preferences;
GSettings *gnome_background_preferences;
GSettings *gnome_media_handling_preferences;
GSettings *gnome_terminal_preferences;
GSettings *cdos_privacy_preferences;
GSettings *cdos_interface_preferences;

GTimeZone      *prefs_current_timezone;
gboolean        prefs_current_24h_time_format;
NemoDateFormat  prefs_current_date_format;

GTimer    *nemo_startup_timer;

gboolean is_special_version;

gboolean default_folder_viewer_changed;

static gboolean ignore_view_metadata = FALSE;
static gboolean inherit_folder_view_preference = FALSE;
static gboolean inherit_show_thumbnails_preference = FALSE;
static int      size_prefixes_preference = 0;

static gchar **file_roller_mimetypes = NULL;


GFileMonitor *tz_mon;

/*
 * Public functions
 */
char *
nemo_global_preferences_get_default_folder_viewer_preference_as_iid (void)
{
	int preference_value;
	const char *viewer_iid;

	preference_value =
		g_settings_get_enum (nemo_preferences, NEMO_PREFERENCES_DEFAULT_FOLDER_VIEWER);

	if (preference_value == NEMO_DEFAULT_FOLDER_VIEWER_LIST_VIEW) {
		viewer_iid = NEMO_LIST_VIEW_IID;
	} else {
		viewer_iid = NEMO_ICON_VIEW_IID;
	}

	return g_strdup (viewer_iid);
}

/* 判断 location 是否为已挂载的光盘目录，获取unix_device 和 relative_path*/
gboolean
nemo_global_preferences_location_is_cd_rom (GFile *location, char **unix_device, char **relative_path)
{
    GVolumeMonitor *monitor;
    GFile *root;
    GList *mounts, *l;
    GVolume *volume;
    gboolean is_cd_rom;
    char *identifier;

    is_cd_rom = FALSE;
    monitor = g_volume_monitor_get ();
    mounts = g_volume_monitor_get_mounts (monitor);

    for (l = mounts; l != NULL; l = l->next) {
        GMount *mount = G_MOUNT (l->data);
        if (g_mount_is_shadowed (mount))
            continue;

        root = g_mount_get_root (mount);

        if (g_file_equal (root, location) || g_file_has_prefix (location, root)) {
            volume = g_mount_get_volume (mount);
            if (volume) {
                identifier = g_volume_get_identifier (volume, G_VOLUME_IDENTIFIER_KIND_UNIX_DEVICE);

                if (identifier && g_str_has_prefix (identifier, "/dev/sr")) {
                    is_cd_rom = TRUE;

                    if (unix_device) {
                        *unix_device = g_volume_get_identifier (volume, G_VOLUME_IDENTIFIER_KIND_UNIX_DEVICE);
                    }

                    if (relative_path) {
                        *relative_path = g_file_get_relative_path (root, location);
                    }
                }
                g_object_unref (volume);
                g_free (identifier);
            }
        }
        g_object_unref (root);

        if (is_cd_rom) {
            break;
        }
    }

    g_list_free_full (mounts, g_object_unref);
    g_object_unref (monitor);

    return is_cd_rom;
}

/* 如果 location 地址 为光盘目录的假地址，获取对应的光盘文件暂存目录，否则返回它本身 */
/* 使点击光盘文件暂存目录显示的假地址时，能够跳转到对应的文件暂存目录中 */
GFile *
nemo_global_preferences_get_burn_cache_location (GFile *location)
{
    GFile *new_location;
    GVolumeMonitor *monitor;
    GFile *root;
    GList *mounts, *l;
    GVolume *volume;
    char *identifier;
    char *relative_path;
    char *real_path;
    char *location_path;

    new_location = NULL;
    location_path = g_file_get_path (location);
    monitor = g_volume_monitor_get ();
    mounts = g_volume_monitor_get_mounts (monitor);

    for (l = mounts; l != NULL; l = l->next) {
        GMount *mount = G_MOUNT (l->data);

        if (g_mount_is_shadowed (mount))
            continue;

        root = g_mount_get_root (mount);
        volume = g_mount_get_volume (mount);
        if (volume) {
            identifier = g_volume_get_identifier (volume, G_VOLUME_IDENTIFIER_KIND_UNIX_DEVICE);
            // 确保location为光盘地址
            if (identifier && g_file_has_prefix (location, root) && g_str_has_prefix (identifier, "/dev/sr")) {
                relative_path = g_file_get_relative_path (root, location);
                real_path = g_build_filename (g_get_user_cache_dir (), "disk-burn", g_strdelimit (identifier, "/", '_'), relative_path, NULL);
                new_location = g_file_new_for_path (real_path);

                g_free (relative_path);
                g_free (real_path);
            }
            // 空白光盘
            if (location_path && identifier && g_str_has_prefix (location_path, "/_dev_sr") && strstr (location_path, g_strdelimit (identifier, "/", '_'))) {
                char *fake_root_path;
                fake_root_path = g_build_filename ("/", identifier, NULL);
                // 点击空白光盘虚拟根目录时进入burn:///界面，而不是地址栏显示 /_dev_sr* 界面
                if (g_strcmp0 (location_path, fake_root_path) == 0) {
                    new_location = g_file_new_for_uri ("burn:///");
                } else {
                    real_path = g_build_filename (g_get_user_cache_dir (), "disk-burn", location_path, NULL);
                    new_location = g_file_new_for_path (real_path);

                    g_free (real_path);
                }
                g_free (fake_root_path);
            }

            g_object_unref (volume);
            g_free (identifier);
        }

        g_object_unref (root);
        if (new_location) {
            break;
        }
    }

    g_list_free_full (mounts, g_object_unref);
    g_object_unref (monitor);
    g_free (location_path);

    return new_location ? new_location : g_object_ref (location);
}

/* 如果 location 是光盘文件暂存目录且对应光盘已挂载，生成基于光盘地址的显示地址 */
gboolean
nemo_global_preferences_location_is_burn_cache_and_cd_rom_mounted (GFile *location, char **new_path)
{
    GVolumeMonitor *monitor;
    GList *mounts, *l;
    GVolume *volume;
    gboolean should_show_burn_bar;
    char *disk_burn_path;
    char *location_path;
    char *identifier;

    should_show_burn_bar = FALSE;

    location_path = g_file_get_path (location);
    disk_burn_path = g_build_filename (g_get_user_cache_dir (), "disk-burn", NULL);
    if (location_path && g_str_has_prefix (location_path, disk_burn_path) && g_strcmp0 (location_path, disk_burn_path) != 0) {
        GFile *disk_burn_location;
        char *relative_path;

        monitor = g_volume_monitor_get ();
        mounts = g_volume_monitor_get_mounts (monitor);
        disk_burn_location = g_file_new_for_path (disk_burn_path);
        relative_path = g_file_get_relative_path (disk_burn_location, location);
        g_object_unref (disk_burn_location);

        for (l = mounts; l != NULL; l = l->next) {
            GMount *mount = G_MOUNT (l->data);
            if (g_mount_is_shadowed (mount))
                continue;

            volume = g_mount_get_volume (mount);
            // 共享目录的 volume 为 NULL
            if (volume) {
                identifier = g_volume_get_identifier (volume, G_VOLUME_IDENTIFIER_KIND_UNIX_DEVICE);
                // 确保 identifier 是 /dev/sr*
                if (relative_path && identifier && g_str_has_prefix (relative_path, g_strdelimit (identifier, "/", '_'))) {
                    should_show_burn_bar = TRUE;

                    if (new_path) {
                        GFile *mount_root; // 光盘挂载点
                        GFile *cache_root; // .../disk-burn/_dev_sr_*
                        char *mount_root_path; // 光盘挂载点 地址
                        char *cache_root_path; // 对应光盘挂载点的暂存区地址
                        char *new_relative_path; 

                        cache_root_path = g_build_filename (disk_burn_path, identifier, NULL);
                        cache_root = g_file_new_for_path (cache_root_path);
                        new_relative_path = g_file_get_relative_path (cache_root, location);

                        mount_root = g_mount_get_root (mount);
                        mount_root_path = g_file_get_path (mount_root);
                        // 空白光盘挂载点地址为空
                        if (mount_root_path) {
                            *new_path = g_build_filename (mount_root_path, new_relative_path, NULL);
                            g_free (mount_root_path);
                        } else {
                            // 空白光盘暂存区目录地址栏显示为 /_dev_sr*
                            *new_path = g_build_filename ("/", relative_path, NULL);
                        }
                        g_object_unref (mount_root);
                        g_object_unref (cache_root);
                        g_free (new_relative_path);
                        g_free (cache_root_path);
                    }
                }

                g_object_unref (volume);
                g_free (identifier);
            }


            if (should_show_burn_bar) {
                break;
            }
        }

        g_list_free_full (mounts, g_object_unref);
        g_object_unref (monitor);
        g_free (relative_path);
    }
    g_free (location_path);
    g_free (disk_burn_path);

    return should_show_burn_bar;
}

char *
nemo_global_preferences_get_home_link_display_name (void)
{
    char *gsettings_value, *display_name;
    gsettings_value = g_settings_get_string (nemo_desktop_preferences, NEMO_PREFERENCES_DESKTOP_HOME_LINK_DISPLAY_NAME);
    if (g_strcmp0 (gsettings_value, "Home") == 0) {
        display_name = g_strdup (_(gsettings_value));
    } else {
        display_name = g_strdup (gsettings_value);
    }
    g_free (gsettings_value);
    return display_name;
}

char *
nemo_global_preferences_get_computer_link_display_name (void)
{
    char *gsettings_value, *display_name;
    gsettings_value = g_settings_get_string (nemo_desktop_preferences, NEMO_PREFERENCES_DESKTOP_COMPUTER_LINK_DISPLAY_NAME);
    if (g_strcmp0 (gsettings_value, "Computer") == 0) {
        display_name = g_strdup (_(gsettings_value));
    } else {
        display_name = g_strdup (gsettings_value);
    }
    g_free (gsettings_value);
    return display_name;
}

char *
nemo_global_preferences_get_trash_link_display_name (void)
{
    char *gsettings_value, *display_name;
    gsettings_value = g_settings_get_string (nemo_desktop_preferences, NEMO_PREFERENCES_DESKTOP_TRASH_LINK_DISPLAY_NAME);
    if (g_strcmp0 (gsettings_value, "Trash") == 0) {
        display_name = g_strdup (_(gsettings_value));
    } else {
        display_name = g_strdup (gsettings_value);
    }
    g_free (gsettings_value);
    return display_name;
}

gboolean
nemo_global_preferences_get_inherit_folder_viewer_preference (void)
{
    return inherit_folder_view_preference;
}

gboolean
nemo_global_preferences_get_ignore_view_metadata (void)
{
    return ignore_view_metadata;
}

gboolean
nemo_global_preferences_get_inherit_show_thumbnails_preference (void)
{
    return inherit_show_thumbnails_preference;
}

int
nemo_global_preferences_get_size_prefix_preference (void)
{
    switch (size_prefixes_preference) {
        case 0: // base-10
            return G_FORMAT_SIZE_DEFAULT;
        case 1: // base-10 full
            return G_FORMAT_SIZE_DEFAULT |                           G_FORMAT_SIZE_LONG_FORMAT;
        case 2: // base-2
            return G_FORMAT_SIZE_DEFAULT | G_FORMAT_SIZE_IEC_UNITS;
        case 3: // base-2 full
            return G_FORMAT_SIZE_DEFAULT | G_FORMAT_SIZE_IEC_UNITS | G_FORMAT_SIZE_LONG_FORMAT;
    }

    return 0;
}

char *
nemo_global_preferences_get_desktop_iid (void)
{
    gboolean use_grid;
    const char *viewer_iid;

    use_grid = g_settings_get_boolean (nemo_desktop_preferences, NEMO_PREFERENCES_USE_DESKTOP_GRID);

    if (use_grid) {
        viewer_iid = NEMO_DESKTOP_ICON_GRID_VIEW_IID;
    } else {
        viewer_iid = NEMO_DESKTOP_ICON_VIEW_IID;
    }

    return g_strdup (viewer_iid);
}

gint
nemo_global_preferences_get_tooltip_flags (void)
{
    NemoFileTooltipFlags flags = NEMO_FILE_TOOLTIP_FLAGS_NONE;

    if (g_settings_get_boolean (nemo_preferences, NEMO_PREFERENCES_TOOLTIP_FILE_TYPE))
        flags |= NEMO_FILE_TOOLTIP_FLAGS_FILE_TYPE;
    if (g_settings_get_boolean (nemo_preferences, NEMO_PREFERENCES_TOOLTIP_MOD_DATE))
        flags |= NEMO_FILE_TOOLTIP_FLAGS_MOD_DATE;
    if (g_settings_get_boolean (nemo_preferences, NEMO_PREFERENCES_TOOLTIP_ACCESS_DATE))
        flags |= NEMO_FILE_TOOLTIP_FLAGS_ACCESS_DATE;
    if (g_settings_get_boolean (nemo_preferences, NEMO_PREFERENCES_TOOLTIP_FULL_PATH))
        flags |= NEMO_FILE_TOOLTIP_FLAGS_PATH;
    if (g_settings_get_boolean (nemo_preferences, NEMO_PREFERENCES_TOOLTIP_CREATED_DATE))
        flags |= NEMO_FILE_TOOLTIP_FLAGS_CREATED_DATE;
    return flags;
}

gboolean
nemo_global_preferences_should_load_plugin (const gchar *name, const gchar *key)
{
    gchar **disabled_list = g_settings_get_strv (nemo_plugin_preferences, key);

    gboolean ret = TRUE;
    guint i = 0;

    for (i = 0; i < g_strv_length (disabled_list); i++) {
        if (g_strcmp0 (disabled_list[i], name) == 0)
            ret = FALSE;
    }

    g_strfreev (disabled_list);
    return ret;
}

/*磁盘重命名时，进入重命名后显示的名称带盘符，需要去掉名称后的盘符*/
char *
nemo_global_preferences_str_delete_drive_letter (const char *string)
{
    char *c;
    char *final;

    g_return_val_if_fail (string != NULL, NULL);

    final = g_strdup (string);

    if (g_str_has_suffix (final, ":)")) {
        for (c = final; *c ; c++) {
            if (*c == ')' && *(c-1) == ':') {
                *(c - 4) = '\0';
            }
        }
    }

    return final;
}

/*磁盘重命名时，限制字符不允许显示，去掉字符串中的限制字符*/
char *
nemo_global_preferences_str_delete_limit_char (const char *string, const gchar *delimiters)
{
    char *c;
    char *p;
    char *final;

    g_return_val_if_fail (string != NULL, NULL);

    final = g_strdup (string);
    if (!delimiters) {
        return final;
    }

    for (c = final; *c; c++)
    {
        if (strchr (delimiters, *c)) {
            for (p = c; *p != '\0'; p++) {
                *p = *(p+1);
            }
        }
    }

    return final;
}

static void
boolean_changed_cb (GSettings *settings,
                    gchar     *key,
                    gboolean  *user_data)
{
    *user_data = g_settings_get_boolean (settings, key);
}

static void
enum_changed_cb (GSettings *settings,
                 gchar     *key,
                 int       *user_data)
{
    *user_data = g_settings_get_enum (settings, key);
}

static void
default_folder_viewer_cb (GSettings *settings,
                          gchar      *key,
                          gpointer   user_data)
{
    g_settings_set_boolean (nemo_preferences, NEMO_PREFERENCES_DEFAULT_FOLDER_VIEWER_CHANGED, TRUE);
}

static void
setup_cached_pref_keys (void)
{
    g_signal_connect (nemo_preferences,
                      "changed::" NEMO_PREFERENCES_IGNORE_VIEW_METADATA,
                      G_CALLBACK (boolean_changed_cb), &ignore_view_metadata);

    boolean_changed_cb (nemo_preferences, NEMO_PREFERENCES_IGNORE_VIEW_METADATA, &ignore_view_metadata);

    g_signal_connect (nemo_preferences,
                      "changed::" NEMO_PREFERENCES_INHERIT_FOLDER_VIEWER,
                      G_CALLBACK (boolean_changed_cb), &inherit_folder_view_preference);

    boolean_changed_cb (nemo_preferences, NEMO_PREFERENCES_INHERIT_FOLDER_VIEWER, &inherit_folder_view_preference);

    g_signal_connect (nemo_preferences,
                      "changed::" NEMO_PREFERENCES_INHERIT_SHOW_THUMBNAILS,
                      G_CALLBACK (boolean_changed_cb), &inherit_show_thumbnails_preference);

    boolean_changed_cb (nemo_preferences, NEMO_PREFERENCES_INHERIT_SHOW_THUMBNAILS, &inherit_show_thumbnails_preference);

    g_signal_connect (nemo_preferences,
                      "changed::" NEMO_PREFERENCES_SIZE_PREFIXES,
                      G_CALLBACK (enum_changed_cb), &size_prefixes_preference);

    enum_changed_cb (nemo_preferences, NEMO_PREFERENCES_SIZE_PREFIXES, &size_prefixes_preference);

    g_signal_connect (nemo_preferences,
                      "changed::" NEMO_PREFERENCES_DEFAULT_FOLDER_VIEWER,
                      G_CALLBACK (default_folder_viewer_cb), NULL);

    g_signal_connect (nemo_preferences,
                      "changed::" NEMO_PREFERENCES_DEFAULT_FOLDER_VIEWER_CHANGED,
                      G_CALLBACK (boolean_changed_cb), &default_folder_viewer_changed);
    
    boolean_changed_cb (nemo_preferences, NEMO_PREFERENCES_DEFAULT_FOLDER_VIEWER_CHANGED, &default_folder_viewer_changed);
}


gchar **
nemo_global_preferences_get_fileroller_mimetypes (void)
{
    static gsize once_init = 0;

    if (g_once_init_enter (&once_init)) {
        if (nemo_is_file_roller_installed ()) {
            GAppInfo *app_info;
            gchar ***results;
            gchar **result;
            gint i;

            results = g_desktop_app_info_search ("file-roller");

            if (results != NULL && results[0] != NULL) {
                const gchar *best;

                best = results[0][0];

                app_info = G_APP_INFO (g_desktop_app_info_new (best));

                if (app_info) {
                    file_roller_mimetypes = g_strdupv ((gchar **) g_app_info_get_supported_types (app_info));
                    g_object_unref (app_info);
                }

                if (app_info == NULL) {
                    g_warning ("Unable to retrieve list of file-roller mimetypes");
                }

                i = 0;
                result = results[i];

                while (result != NULL) {
                    g_strfreev (result);
                    result = results[++i];
                }

                g_free (results);
            }
        }

        g_once_init_leave (&once_init, 1);
    }

    return file_roller_mimetypes;
}

static void
on_time_data_changed (gpointer user_data)
{
    prefs_current_date_format = g_settings_get_enum (nemo_preferences, NEMO_PREFERENCES_DATE_FORMAT);
    prefs_current_24h_time_format = g_settings_get_boolean (cdos_interface_preferences, "clock-use-24h");

    if (prefs_current_timezone != NULL) {
        g_time_zone_unref (prefs_current_timezone);
    }

    prefs_current_timezone = g_time_zone_new_local ();
}

static void
setup_cached_time_data (void)
{
    GFile *tz;

    prefs_current_timezone = NULL;

    g_signal_connect_swapped (nemo_preferences,
                              "changed::" NEMO_PREFERENCES_DATE_FORMAT,
                              G_CALLBACK (on_time_data_changed), NULL);

    g_signal_connect_swapped (cdos_interface_preferences,
                              "changed::clock-use-24h",
                              G_CALLBACK (on_time_data_changed), NULL);


    tz = g_file_new_for_path ("/etc/localtime");

    tz_mon = g_file_monitor_file (tz, 0, NULL, NULL);
    g_object_unref (tz);

    g_signal_connect_swapped (tz_mon,
                              "changed",
                              G_CALLBACK (on_time_data_changed), NULL);

    on_time_data_changed (NULL);
}

static gboolean
nemo_global_preferences_is_special_version (void)
{
    gboolean selinux_enabled;

    selinux_enabled = is_selinux_enabled ();

    return selinux_enabled;
}

void
nemo_global_preferences_init (void)
{
	static gboolean initialized = FALSE;

	if (initialized) {
		return;
	}

	initialized = TRUE;

	nemo_preferences = g_settings_new("org.nemo.preferences");
	nemo_window_state = g_settings_new("org.nemo.window-state");
	nemo_icon_view_preferences = g_settings_new("org.nemo.icon-view");
	nemo_list_view_preferences = g_settings_new("org.nemo.list-view");
	nemo_compact_view_preferences = g_settings_new("org.nemo.compact-view");
	nemo_desktop_preferences = g_settings_new("org.nemo.desktop");
	nemo_tree_sidebar_preferences = g_settings_new("org.nemo.sidebar-panels.tree");
    nemo_plugin_preferences = g_settings_new("org.nemo.plugins");
    nemo_menu_config_preferences = g_settings_new("org.nemo.preferences.menu-config");
	gnome_lockdown_preferences = g_settings_new("org.cdos.desktop.lockdown");
	gnome_background_preferences = g_settings_new("org.cdos.desktop.background");
	gnome_media_handling_preferences = g_settings_new("org.cdos.desktop.media-handling");
	gnome_terminal_preferences = g_settings_new("org.cdos.desktop.default-applications.terminal");
    cdos_privacy_preferences = g_settings_new("org.cdos.desktop.privacy");
	cdos_interface_preferences = g_settings_new ("org.cdos.desktop.interface");

    is_special_version = nemo_global_preferences_is_special_version ();
    //禁用ZD代码
    is_special_version = FALSE;

    setup_cached_pref_keys ();
    setup_cached_time_data ();

    eel_debug_call_at_shutdown (nemo_global_preferences_finalize);
}

void
nemo_global_preferences_finalize (void)
{
    g_strfreev (file_roller_mimetypes);
    g_object_unref (tz_mon);

    g_object_unref (nemo_preferences);
    g_object_unref (nemo_window_state);
    g_object_unref (nemo_icon_view_preferences);
    g_object_unref (nemo_list_view_preferences);
    g_object_unref (nemo_compact_view_preferences);
    g_object_unref (nemo_desktop_preferences);
    g_object_unref (nemo_tree_sidebar_preferences);
    g_object_unref (nemo_plugin_preferences);
    g_object_unref (nemo_menu_config_preferences);
    g_object_unref (gnome_lockdown_preferences);
    g_object_unref (gnome_background_preferences);
    g_object_unref (gnome_media_handling_preferences);
    g_object_unref (gnome_terminal_preferences);
    g_object_unref (cdos_privacy_preferences);
    g_object_unref (cdos_interface_preferences);
}
