#include "nemo-smb-data.h"

static guint save_in_idle_source_id = 0;

static gchar *
get_keyfile_path (void)
{
    return g_build_filename (g_get_user_data_dir (), "smb-automounts", NULL);
}

static GKeyFile *
load_smb_keyfile (void)
{
    GKeyFile *retval;
    GError *error = NULL;
    gchar *filename;

    retval = g_key_file_new ();
    filename = get_keyfile_path ();

    g_key_file_load_from_file (retval,
                               filename,
                               G_KEY_FILE_NONE,
                               &error);

    if (error != NULL)
    {
        if (!g_error_matches (error,
                              G_FILE_ERROR,
                              G_FILE_ERROR_NOENT))
        {
            g_print ("Unable to open the smb keyfile: %s\n", error->message);
        }

        g_error_free (error);
    }

    g_free (filename);

    return retval;
}

static GKeyFile *
get_keyfile (void)
{
    static gboolean keyfile_loaded = FALSE;
    static GKeyFile *keyfile = NULL;

    if (!keyfile_loaded)
    {
        keyfile = load_smb_keyfile ();
        keyfile_loaded = TRUE;
    }

    return keyfile;
}

static gboolean
save_in_idle_cb (gpointer data)
{
    GKeyFile *keyfile = data;
    gchar *contents, *filename;
    gsize length;
    GError *error = NULL;

    save_in_idle_source_id = 0;

    contents = g_key_file_to_data (keyfile, &length, NULL);
    filename = get_keyfile_path ();

    if (contents != NULL) {
        g_file_set_contents (filename,
                             contents, length,
                             &error);
        g_free (contents);
    }

    if (error != NULL) {
        g_warning ("Couldn't save the smb keyfile to disk: %s",
                   error->message);
        g_error_free (error);
    }

    g_free (filename);

    return FALSE;
}

static void
save_in_idle (GKeyFile *keyfile)
{
    if (save_in_idle_source_id != 0) {
        g_source_remove (save_in_idle_source_id);
    }

    save_in_idle_source_id = g_idle_add (save_in_idle_cb, keyfile);
}

static gchar *
remove_tail_left_slash (const gchar *string)
{
    gsize len;
    gchar *res;

    if (string == NULL) {
        return NULL;
    }

    res = g_strdup (string);
    len = strlen (string);

    if (res[len - 1] == '/') {
        res[len - 1] = '\0';
    }

    return res;
}

void
smb_conf_set_uri (gchar *uri)
{
    GKeyFile *keyfile;
    gchar **uris;
    gchar *new_uri;
    gsize length;
    GPtrArray *array;
    gint i;

    keyfile = get_keyfile ();
    array = g_ptr_array_new ();
    new_uri = remove_tail_left_slash (uri);

    uris = g_key_file_get_string_list (keyfile, "smb-uris", "uris", &length, NULL);

    if (uris == NULL) {
        g_ptr_array_add (array, g_strdup (new_uri));
    } else {
        if (g_strv_contains ((const gchar * const *) uris, new_uri)) {
            return;
        }

        for (i = 0; uris[i] != NULL; i++) {
            g_ptr_array_add (array, g_strdup (uris[i]));
        }

        g_ptr_array_add (array, g_strdup (new_uri));
    }

    g_ptr_array_add (array, NULL);

    g_strfreev (uris);
    uris = (gchar **) g_ptr_array_free (array, FALSE);

    g_key_file_set_string_list (keyfile, "smb-uris", "uris", (const gchar * const *) uris, g_strv_length (uris));

    save_in_idle (keyfile);

    g_strfreev (uris);
    g_free (new_uri);
}