#include <glib.h>
#include <gio/gio.h>

static int outstanding_mounts = 0;
static GMainLoop *loop;

static gchar *
get_keyfile_path (void)
{
    return g_build_filename (g_get_user_data_dir (), "smb-automounts", NULL);
}

static GKeyFile *
load_smb_keyfile (void)
{
    GKeyFile *retval;
    GError *error = NULL;
    gchar *filename;

    retval = g_key_file_new ();
    filename = get_keyfile_path ();

    g_key_file_load_from_file (retval,
                    filename,
                    G_KEY_FILE_NONE,
                    &error);

    if (error != NULL) {
        if (!g_error_matches (error,
                        G_FILE_ERROR,
                        G_FILE_ERROR_NOENT)) {
            g_print ("Unable to open the smb keyfile: %s\n", error->message);
        }

        g_error_free (error);
    }

    g_free (filename);

    return retval;
}

static GKeyFile *
get_keyfile (void)
{
    static gboolean keyfile_loaded = FALSE;
    static GKeyFile *keyfile = NULL;

    if (!keyfile_loaded) {
        keyfile = load_smb_keyfile ();
        keyfile_loaded = TRUE;
    }

    return keyfile;
}

static void
mount_done_cb (GObject *source_object,
               GAsyncResult *res,
               gpointer user_data)
{
    gboolean succeeded;
    char *uri;
    GError *error = NULL;

    succeeded = g_file_mount_enclosing_volume_finish (G_FILE (source_object), res, &error);

    if (!succeeded) {
        uri = g_file_get_uri (G_FILE (source_object));

        switch (error->code)
        {
            case G_IO_ERROR_ALREADY_MOUNTED:
                g_print ("%s already mounted\n", uri);
                break;
            case G_IO_ERROR_FAILED:
                g_print ("Mount saved smb \"%s\" failed\n", uri);
                break;
            
            default:
                break;
        }

        g_free (uri);
        g_error_free (error);
    }

    outstanding_mounts--;

    if (outstanding_mounts == 0)
        g_main_loop_quit (loop);
}

static gboolean
idle_callback (gpointer user_data)
{
    g_main_loop_quit (loop);

    return FALSE;
}

static void
mount_smb (void)
{
    GKeyFile *keyfile;
    gchar **uris;
    gsize length;
    gint i;
    GFile *file;
    GMountOperation *mount_op;

    keyfile = get_keyfile ();
    uris = g_key_file_get_string_list (keyfile, "smb-uris", "uris", &length, NULL);

    mount_op = g_mount_operation_new ();

    if (uris) {
        for (i = 0; uris[i] != NULL; i++)
        {
            file = g_file_new_for_uri (uris[i]);

            g_file_mount_enclosing_volume (file,
                                           0,
                                           mount_op,
                                           NULL,
                                           mount_done_cb,
                                           NULL);

            outstanding_mounts++;

            g_object_unref (file);
        }
    }

    g_object_unref (mount_op);

    if (uris == NULL) {
        g_idle_add ((GSourceFunc) idle_callback, NULL);
    }
}

int main (int argc, char *argv[])
{
    loop = g_main_loop_new (NULL, FALSE);

    mount_smb ();

    g_main_loop_run (loop);

    if (loop != NULL) {
        g_main_loop_unref (loop);
    }

    return 0;
}