/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
* Copyright (C) 2006 Paolo Borelli <pborelli@katamail.com>
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street - Suite 500, Boston, MA 02110-1335, USA.
*
* Authors: Paolo Borelli <pborelli@katamail.com>
*
*/

#include "config.h"

#include <glib/gstdio.h>
#include <glib/gi18n.h>
#include <gtk/gtk.h>
#include <disk-burn/disk-burn-window.h>

#include "nemo-burn-bar.h"

#include "nemo-view.h"
#include <libnemo-private/nemo-global-preferences.h>
#include <libnemo-private/nemo-file.h>
#include <eel/eel-stock-dialogs.h>

#define NEMO_BURN_BAR_GET_PRIVATE(o)\
    (G_TYPE_INSTANCE_GET_PRIVATE ((o), NEMO_TYPE_BURN_BAR, NemoBurnBarPrivate))

enum {
    PROP_VIEW = 1,
    NUM_PROPERTIES
};

enum {
    BURN_TO_DISK = 1
};

struct NemoBurnBarPrivate
{
    NemoView *view;
};

G_DEFINE_TYPE (NemoBurnBar, nemo_burn_bar, GTK_TYPE_INFO_BAR);


static void
nemo_burn_bar_set_property (GObject      *object,
                guint         prop_id,
                const GValue *value,
                GParamSpec   *pspec)
{
    NemoBurnBar *bar;

    bar = NEMO_BURN_BAR (object);

    switch (prop_id) {
    case PROP_VIEW:
        bar->priv->view = g_value_get_object (value);
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
        break;
    }
}

static void
nemo_burn_bar_get_property (GObject    *object,
             guint       prop_id,
             GValue     *value,
             GParamSpec *pspec)
{
    NemoBurnBar *bar;

    bar = NEMO_BURN_BAR (object);

    switch (prop_id) {
    case PROP_VIEW:
        g_value_set_object (value, bar->priv->view);
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
        break;
    }
}

static void
nemo_burn_bar_dispose (GObject *obj)
{
    G_OBJECT_CLASS (nemo_burn_bar_parent_class)->dispose (obj);
}

static void
nemo_burn_bar_class_init (NemoBurnBarClass *klass)
{
    GObjectClass *object_class;

    object_class = G_OBJECT_CLASS (klass);

    object_class->set_property = nemo_burn_bar_set_property;
    object_class->get_property = nemo_burn_bar_get_property;
    object_class->dispose = nemo_burn_bar_dispose;

    g_object_class_install_property (object_class,
                    PROP_VIEW,
                    g_param_spec_object ("view",
                                "view",
                                "the NemoView",
                                NEMO_TYPE_VIEW,
                                G_PARAM_WRITABLE |
                                G_PARAM_CONSTRUCT_ONLY |
                                G_PARAM_STATIC_STRINGS));

    g_type_class_add_private (klass, sizeof (NemoBurnBarPrivate));
}

gboolean
get_mount_point_and_unix_device (GFile *location, char **unix_device, char **mount_point)
{
    GVolumeMonitor *monitor;
    GFile *root, *disk_burn_location;
    GList *mounts, *l;
    GVolume *volume;
    gboolean is_cd_rom;
    char *identifier;
    char *location_path;
    char *disk_burn_path;
    char *relative_path;

    is_cd_rom = FALSE;
    identifier = NULL;
    relative_path = NULL;
    location_path = NULL;

    // burn bar只显示在光盘目录和光盘文件暂存目录，location只可能是光盘中文件或暂存目录中文件
    location_path = g_file_get_path (location);
    disk_burn_path = g_build_filename (g_get_user_cache_dir (), "disk-burn", NULL);
    if (location_path && g_str_has_prefix (location_path, disk_burn_path)) {
        disk_burn_location = g_file_new_for_path (disk_burn_path);
        relative_path = g_file_get_relative_path (disk_burn_location, location);
        g_object_unref (disk_burn_location);
    }

    monitor = g_volume_monitor_get ();
    mounts = g_volume_monitor_get_mounts (monitor);

    for (l = mounts; l != NULL; l = l->next) {
        GMount *mount = G_MOUNT (l->data);
        if (g_mount_is_shadowed (mount))
            continue;

        root = g_mount_get_root (mount);

        // 当前目录为光盘文件暂存目录或光盘根目录或光盘子目录
        if (relative_path || g_file_equal (root, location) || g_file_has_prefix (location, root)) {
            volume = g_mount_get_volume (mount);
            if (volume) {
                identifier = g_volume_get_identifier (volume, G_VOLUME_IDENTIFIER_KIND_UNIX_DEVICE);

                // relative_path 不为 NULL， 说明当前目录是光盘文件暂存目录，获取与当前地址匹配的unix_device
                if (identifier && relative_path && g_str_has_prefix (relative_path, g_strdelimit (identifier, "/", '_'))) {
                    is_cd_rom = TRUE;
                    if (unix_device) {
                        *unix_device = g_volume_get_identifier (volume, G_VOLUME_IDENTIFIER_KIND_UNIX_DEVICE);
                    }
                    if (mount_point) {
                        *mount_point = g_file_get_path (root);
                    }
                } else if (identifier && g_str_has_prefix (identifier, "/dev/sr")) {
                    is_cd_rom = TRUE;
                    if (unix_device) {
                        *unix_device = g_volume_get_identifier (volume, G_VOLUME_IDENTIFIER_KIND_UNIX_DEVICE);
                    }
                    if (mount_point) {
                        *mount_point = g_file_get_path (root);
                    }
                }
                g_object_unref (volume);
                g_free (identifier);
            }
        }
        g_object_unref (root);

        if (is_cd_rom) {
            break;
        }
    }

    g_list_free_full (mounts, g_object_unref);
    g_object_unref (monitor);
    g_free (disk_burn_path);
    g_free (location_path);
    g_free (relative_path);

    return is_cd_rom;
}

/* 检查光盘缓存目录中是否存在需要刻录的内容 */
static gboolean
cd_should_burn (GFile *file, GFile *cache_root, char *unix_device, char *mount_point)
{
    GFileInfo *info;
    GFileEnumerator *enumerator;
    GFile *child;

    char *relative_path;
    char *check_path;
    gboolean ret;

    ret = FALSE;
    if (g_file_query_file_type (file, G_FILE_QUERY_INFO_NONE, NULL) == G_FILE_TYPE_DIRECTORY) {
        enumerator = g_file_enumerate_children (file, G_FILE_ATTRIBUTE_STANDARD_SIZE","G_FILE_ATTRIBUTE_STANDARD_TYPE","G_FILE_ATTRIBUTE_STANDARD_NAME, 0, NULL, NULL);
        if (enumerator != NULL) {
            while ((info = g_file_enumerator_next_file (enumerator, NULL, NULL))) {
                child = g_file_get_child (file,
                            g_file_info_get_name (info));

                if (mount_point) {
                    // 根据文件地址生成对应刻到光盘的地址，判断这个地址的文件是否存在
                    // 存在则说明在光盘上已存在这个文件，不存在则需要进行刻录
                    relative_path = g_file_get_relative_path (cache_root, child);
                    check_path = g_build_filename (mount_point, relative_path, NULL);
                    if (!g_file_test (check_path, G_FILE_TEST_EXISTS)) {
                        ret = TRUE;
                    } else if (g_file_info_get_file_type (info) == G_FILE_TYPE_DIRECTORY) {
                        ret = cd_should_burn (child, cache_root, unix_device, mount_point);
                    }
                    g_free (relative_path);
                    g_free (check_path);
                } else {
                    // mount_point == NULL, 说明光盘为空白光盘，
                    // 程序执行到这里，说明暂存区中有文件，需要刻录
                    ret = TRUE;
                }


                g_object_unref (child);
                g_object_unref (info);

                if (ret) {
                    break;
                }
            }
            g_file_enumerator_close (enumerator, NULL, NULL);
            g_object_unref (enumerator);
        }
    }
    return ret;
}

static void
burn_bar_response_cb (GtkInfoBar *infobar,
            gint response_id,
            gpointer user_data)
{
    NemoBurnBar *bar;
    GtkWidget *window;
    NemoFile *file;
    GFile *location;
    char *unix_device;
    char *mount_point;
    char *burn_cache_name;
    char *cache_root_path;

    bar = NEMO_BURN_BAR (infobar);
    window = gtk_widget_get_toplevel (GTK_WIDGET (bar));

    file = nemo_view_get_directory_as_file (bar->priv->view);
    location = nemo_file_get_location (file);
    unix_device = NULL;
    mount_point = NULL;

    get_mount_point_and_unix_device (location, &unix_device, &mount_point);
    burn_cache_name = g_strdup (unix_device);
    cache_root_path = g_build_filename (g_get_user_cache_dir (), "disk-burn", g_strdelimit (burn_cache_name, "/", '_'), NULL);
    g_free (burn_cache_name);


    switch (response_id) {
    case BURN_TO_DISK:
        if (g_file_test (cache_root_path, G_FILE_TEST_EXISTS)) {
            GFile *cache_root_location;
            cache_root_location = g_file_new_for_path (cache_root_path);
            if (!cd_should_burn (cache_root_location, cache_root_location, unix_device, mount_point)) {
                eel_show_error_dialog (_("Tips"), _("No files need to be burned"), GTK_WINDOW (window));
            } else {
                disk_burn_window_new (unix_device, cache_root_path);
            }
            g_object_unref (cache_root_location);
        } else {
            eel_show_error_dialog (_("Tips"), _("No files need to be burned"), GTK_WINDOW (window));
        }
        break;
    default:
        break;
    }
    g_object_unref (location);
    g_free (cache_root_path);
    g_free (mount_point);
    g_free (unix_device);
}

static void
nemo_burn_bar_init (NemoBurnBar *bar)
{
    GtkWidget *action_area, *w;

    bar->priv = NEMO_BURN_BAR_GET_PRIVATE (bar);
    action_area = gtk_info_bar_get_action_area (GTK_INFO_BAR (bar));

    gtk_orientable_set_orientation (GTK_ORIENTABLE (action_area),
                    GTK_ORIENTATION_HORIZONTAL);

    gtk_style_context_add_class (gtk_widget_get_style_context (GTK_WIDGET (bar)),
                    "nemo-burn");

    w = gtk_info_bar_add_button (GTK_INFO_BAR (bar),
                    _("Burn"),
                    BURN_TO_DISK);
    gtk_widget_set_tooltip_text (w,
                    _("Burn to disk"));

    g_signal_connect (bar, "response",
            G_CALLBACK (burn_bar_response_cb), bar);
}

GtkWidget *
nemo_burn_bar_new (NemoView *view)
{
    return g_object_new (NEMO_TYPE_BURN_BAR,
                "view", view,
                NULL);
}
